/* $AoiMoe: gesture_conf.cc,v 1.4 2003/02/12 07:51:06 tshiozak Exp $ */

/*-
 * Copyright (c)2003 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <iostream>
#include <string>
#include <list>
#include <map>
#include <algorithm>
#include <exception>
#include <string.h>
#include <ctype.h>

#ifdef __GNUC__
#pragma implementation "gesture_conf.hh"
#endif

#include "gesture_conf_p.h"
#include "gesture_conf.hh"

using namespace std;

#if 0
#define LEXFREE(ptr)					\
do {							\
	cerr << "lexfree: \"" << (char *) ptr << "\":"	\
	     << (void *)ptr << endl;			\
	free((void *)ptr);				\
} while (/*CONSTCOND*/0)
#else
#define LEXFREE(ptr) free((void *)ptr);
#endif

extern "C" { void yyparse(); }

namespace ParseManager_D_
{

class ParseManagerImpl_C : public Impl_I
{
	template <class T_>
	class Stack_
	{
		typedef typename T_::Handle Handle_;
		Handle_ m_ptr_;
		void create()
		{
			if (m_ptr_==NULL)
				m_ptr_ = new T_;
		}
	public:
		Stack_() : m_ptr_(NULL) { }
		~Stack_() { }
		bool isnull() const { return (m_ptr_==NULL); }
		Handle_ release()
		{
			this->create();
			Handle_ ret = m_ptr_;
			m_ptr_ = NULL;
			return (ret);
		}
		Handle_ operator -> ()
		{
			this->create();
			return (m_ptr_);
		}
	};
	// stacks
	Stack_<Stroke_C> m_currentStroke_;
	Stack_<Sequence_C> m_currentSequence_;
	Stack_<MotionSet_C> m_currentMotionSet_;
	// pools
	MotionSetPool_C::Handle m_motionSetPool_;
	TargetPool_C::Handle m_targetPool_;
	//
	KeysymConverter_I::Handle m_keysymConverter_;
	ErrorMessageProc_I::Handle m_errorMessageProc_;
	//
	std::string m_gestureConf_;
	std::string::const_iterator m_gestureIterator_;
	int m_line_;
public:
	static ParseManagerImpl_C *current;
	ParseManagerImpl_C(KeysymConverter_I::Handle conv,
			   ErrorMessageProc_I::Handle err)
		: m_motionSetPool_(new MotionSetPool_C),
		  m_targetPool_(new TargetPool_C),
		  m_keysymConverter_(conv),
		  m_errorMessageProc_(err)
	{
	}
	~ParseManagerImpl_C() { }
	void PushKeyElementToCurrentStroke(std::string const &keysym)
	{
		m_currentStroke_->PushKeyCode((*m_keysymConverter_)(keysym));
	}
	void PushCurrentStrokesToCurrentSequence()
	{
		m_currentSequence_->PushStroke(m_currentStroke_.release());
	}
	void PushMotionToCurrentMotionSet(std::string const &motion,
					  std::string const &title)
	{
		m_currentMotionSet_->PushMotion(motion, title,
						m_currentSequence_.release());
	}
	void ImportMotionSetToCurrentMotionSet(std::string const &src)
	{
		MotionSet_C::Handle m = m_motionSetPool_->Lookup(src);
		m_currentMotionSet_->ImportMotionSet(m);
	}
	void StoreNamedMotionSetToMotionSetPool(std::string const &name)
	{
		m_motionSetPool_->PushMotionSet(
			name, m_currentMotionSet_.release());
	}
	void CreateNewTargetFromCurrentMotionSet(std::string const &appname,
						 std::string const &title)
	{
		m_targetPool_->PushTarget(appname, title,
					  m_currentMotionSet_.release());
	}
	void CreateNewTargetFromNamedMotionSet(std::string const &appname,
					       std::string const &title,
					       std::string const &motion)
	{
		MotionSet_C::Handle m = m_motionSetPool_->Lookup(motion);
		m_targetPool_->PushTarget(appname, title, m);
	}
	int GetGesturesChar(void)
	{
		if (m_gestureIterator_ == m_gestureConf_.end())
			return (EOF);
		return ((unsigned char)*m_gestureIterator_++);
	}
	void ProcGesturesError(std::string const &err,
			       std::string const &token)
	{
		(*m_errorMessageProc_)(m_line_, err, token);
	}
	void IncrGestureLine()
	{
		m_line_++;
	}
	//
	TargetPool_C::Handle GetTargetPool() const
	{
		return (m_targetPool_);
	}
	void Parse(std::string const &gesture)
	{
		this->current = this;
		m_gestureConf_ = gesture;
		m_gestureIterator_ = m_gestureConf_.begin();
		m_line_ = 1;
		yyparse();
	}
};

Impl_I *
Impl_I::create(KeysymConverter_I::Handle conv, ErrorMessageProc_I::Handle err)
{
	return (new ParseManagerImpl_C(conv, err));
}


ParseManagerImpl_C *ParseManagerImpl_C::current;

ParseManager_D_::Impl_I::~Impl_I() { }

} // namespace ParseManager_D_


KeysymConverter_I::~KeysymConverter_I() { }
ErrorMessageProc_I::~ErrorMessageProc_I() { }


// C <-> C++ bridge

static inline ParseManager_D_::ParseManagerImpl_C &
pmcur()
{
	return (*ParseManager_D_::ParseManagerImpl_C::current);
}

void PushStroke(char *keysym)
{
	try {
		pmcur().PushKeyElementToCurrentStroke(keysym);
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
	LEXFREE(keysym);
}
void PushSequence(void)
{
	try {
		pmcur().PushCurrentStrokesToCurrentSequence();
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
}
void PushNewMotion(char *motion, char *title)
{
	try {
		if (title==NULL)
			title = strdup(motion);
		pmcur().PushMotionToCurrentMotionSet(motion, title);
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
	LEXFREE(motion);
	LEXFREE(title);
}
void CreateNamedMotionSetFromStack(char *name)
{
	try {
		pmcur().StoreNamedMotionSetToMotionSetPool(name);
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
	LEXFREE(name);
}
void ImportMotion(char *name)
{
	try {
		pmcur().ImportMotionSetToCurrentMotionSet(name);
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
	LEXFREE(name);
}
void CreateNewTargetFromStack(char *appname, char *title)
{
	try {
		if (title==NULL)
			title = strdup(appname);
		pmcur().CreateNewTargetFromCurrentMotionSet(appname, title);
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
	LEXFREE(appname);
	LEXFREE(title);
}
void CreateNewTargetFromNamedMotionSet(char *appname, char *title,
				       char *motion)
{
	try {
		if (title==NULL)
			title = strdup(appname);
		pmcur().CreateNewTargetFromNamedMotionSet(appname, title,
							  motion);
	}
	catch (Base_E const &ex) {
		cerr << ex.what() << endl;
	}
	LEXFREE(appname);
	LEXFREE(title);
	LEXFREE(motion);
}

int GetGesturesChar(void)
{
	return (pmcur().GetGesturesChar());
}

void ProcGesturesError(const char *err, const char *token)
{
	pmcur().ProcGesturesError(err, token);
}

void IncrGestureLine()
{
	pmcur().IncrGestureLine();
}
