/* $AoiMoe: lex.l,v 1.5 2003/02/06 13:34:17 tshiozak Exp $ */

/*-
 * Copyright (c)2003 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

%{
#include <string.h>
#include "gesture_conf_p.h"
#include "yacc.h"
#define YY_INPUT(buf, result, max_size)		\
{						\
	int c = GetGesturesChar();		\
	if (c!=EOF) {				\
		buf[0] = c;			\
		result = 1;			\
	} else					\
		result = YY_NULL;		\
}
#if 0
#define DPR(x) printf x
#else
#define DPR(x)
#endif
%}

%s motion
%s keysym
%s import

KEYSYMCHARS	[a-zA-Z_]
ATOMCHARS	[a-zA-Z_]
MOTIONCHARS	[LlRrUuDd]
DEFM		[Dd][Ee][Ff][Mm]
TARGET		[Tt][Aa][Rr][Gg][Ee][Tt]

%%

[\t ]+			;
\r\n|\n|\r		IncrGestureLine();

#.*			;

{DEFM}			{ DPR(("DEFM\n")); return (T_DEFM); }
{TARGET}		{ DPR(("TARGET\n")); return (T_TARGET); }
<motion>\<		{ DPR(("<\n")); BEGIN(import); return ('<'); }
<import>\>		{ DPR((">\n")); BEGIN(motion); return ('>'); }
\{			{ DPR(("{\n")); BEGIN(motion); return ('{'); }
<motion>\}		{ DPR(("}\n")); BEGIN(INITIAL); return ('}'); }
<motion>\[		{ DPR(("[\n")); BEGIN(keysym); return ('['); }
<keysym>\]		{ DPR(("]\n")); BEGIN(motion); return (']'); }
\;			{ DPR((";\n")); return (';'); }

\"[^\"]*\"		{
				char *end;
				yylval.str = yytext+1;
				for (end=yytext+1; *end != '\"'; end++)
					;
				*end = '\0';
				yylval.str = strdup(yylval.str);
				DPR(("STRING=%s\n", yylval.str)); 
				return (T_STRING);
			}

<keysym>{KEYSYMCHARS}+	{
				yylval.str = yytext;
				yylval.str = strdup(yylval.str);
				DPR(("KEYSYM=%s\n", yylval.str));
				return (T_KEYSYM);
			}


<motion>{MOTIONCHARS}+	{ 
				yylval.str = yytext;
				yylval.str = strdup(yylval.str);
				DPR(("MOTION=%s\n", yylval.str));  
				return (T_MOTION); 
			}

{ATOMCHARS}+		{
				yylval.str = yytext;
				yylval.str = strdup(yylval.str);
				DPR(("ATOM=%s\n", yylval.str));
				return (T_ATOM);
			}

.			{
				yylval.str = yytext;
				yylval.str = strdup(yylval.str);
				DPR(("GARBAGE=%s\n", yylval.str));  
				return (T_GARBAGE); 
			}

%%

#ifndef yywrap
int
yywrap()
{
	return (1);
}
#endif
